from dvc.parsing import DataResolver


def test_params(tmp_dir, dvc):
    (tmp_dir / "params.yaml").dump(
        {"params": {"param1": "params.json", "param2": "params.toml"}}
    )

    template = {"params": ["${params.param1}", "param11", "${params.param2}"]}
    resolver = DataResolver(dvc, tmp_dir, template)
    assert resolver.resolve_params() == ["params.json", "param11", "params.toml"]


def test_metrics(tmp_dir, dvc):
    (tmp_dir / "params.yaml").dump(
        {"metrics": {"metric1": "metrics.json", "metric2": "metrics.toml"}}
    )

    template = {"metrics": ["${metrics.metric1}", "metric11", "${metrics.metric2}"]}
    resolver = DataResolver(dvc, tmp_dir, template)
    assert resolver.resolve_metrics() == ["metrics.json", "metric11", "metrics.toml"]


def test_plots(tmp_dir, dvc):
    template = {
        "plots": [
            {
                "${plots.plot1_name}": {
                    "x": "${plots.x_cls}",
                    "y": {
                        "train_classes.csv": "${plots.y_train_cls}",
                        "test_classes.csv": [
                            "${plots.y_train_cls}",
                            "${plots.y_test_cls}",
                        ],
                    },
                    "title": "Compare test vs train confusion matrix",
                    "template": "confusion",
                    "x_label": "Actual class",
                    "y_label": "Predicted class",
                }
            },
            {"eval/importance2.png": None},
            {"${plots.plot3_name}": None},
            "eval/importance4.png",
            "${plots.plot5_name}",
        ],
    }

    (tmp_dir / "params.yaml").dump(
        {
            "plots": {
                "x_cls": "actual_class",
                "y_train_cls": "predicted_class",
                "y_test_cls": "predicted_class2",
                "plot1_name": "eval/importance1.png",
                "plot3_name": "eval/importance3.png",
                "plot5_name": "eval/importance5.png",
            }
        }
    )
    resolver = DataResolver(dvc, tmp_dir, template)
    assert resolver.resolve_plots() == [
        {
            "eval/importance1.png": {
                "x": "actual_class",
                "y": {
                    "train_classes.csv": "predicted_class",
                    "test_classes.csv": ["predicted_class", "predicted_class2"],
                },
                "title": "Compare test vs train confusion matrix",
                "template": "confusion",
                "x_label": "Actual class",
                "y_label": "Predicted class",
            }
        },
        {"eval/importance2.png": None},
        {"eval/importance3.png": None},
        "eval/importance4.png",
        "eval/importance5.png",
    ]


def test_artifacts(tmp_dir, dvc):
    template = {
        "artifacts": {
            "${artifacts.name}": {
                "path": "${artifacts.path}",
                "type": "model",
                "desc": "CV classification model, ResNet50",
                "labels": ["${artifacts.label1}", "${artifacts.label2}"],
                "meta": {"framework": "${artifacts.framework}"},
            }
        }
    }

    (tmp_dir / "params.yaml").dump(
        {
            "artifacts": {
                "name": "cv-classification",
                "path": "models/resnet.pt",
                "label1": "resnet50",
                "label2": "classification",
                "framework": "pytorch",
            }
        }
    )

    resolver = DataResolver(dvc, tmp_dir, template)
    assert resolver.resolve_artifacts() == {
        "cv-classification": {
            "path": "models/resnet.pt",
            "type": "model",
            "desc": "CV classification model, ResNet50",
            "labels": ["resnet50", "classification"],
            "meta": {"framework": "pytorch"},
        }
    }


def test_datasets(tmp_dir, dvc):
    template = {
        "datasets": [
            {"name": "${ds1.name}", "url": "${ds1.url}", "type": "dvcx"},
            {
                "name": "${ds2.name}",
                "url": "${ds2.url}",
                "type": "dvc",
                "path": "${ds2.path}",
            },
            {
                "name": "${ds3.name}",
                "url": "${ds3.url}",
                "type": "url",
            },
        ]
    }

    (tmp_dir / "params.yaml").dump(
        {
            "ds1": {"name": "dogs", "url": "dvcx://dogs"},
            "ds2": {
                "name": "example-get-started",
                "url": "git@github.com:iterative/example-get-started.git",
                "path": "path",
            },
            "ds3": {
                "name": "cloud-versioning-demo",
                "url": "s3://cloud-versioning-demo",
            },
        }
    )

    resolver = DataResolver(dvc, tmp_dir, template)
    assert resolver.resolve_datasets() == [
        {"name": "dogs", "url": "dvcx://dogs", "type": "dvcx"},
        {
            "name": "example-get-started",
            "url": "git@github.com:iterative/example-get-started.git",
            "type": "dvc",
            "path": "path",
        },
        {
            "name": "cloud-versioning-demo",
            "url": "s3://cloud-versioning-demo",
            "type": "url",
        },
    ]
